# CZTile provides a set of tiling strategies
# Copyright 2022 Carl Zeiss Microscopy GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# To obtain a commercial version please contact Carl Zeiss Microscopy GmbH.
"""Contains all tiling strategies supported by cztile"""
from dataclasses import dataclass
from typing import List, NamedTuple, Union, Tuple
from abc import abstractmethod, ABCMeta

Rectangle = NamedTuple("Rectangle", [("x", int), ("y", int), ("w", int), ("h", int)])


@dataclass
class Tile1d:
    """A single 1D tile.

    Attributes:
        left_most_tile_pixel: the left-most pixel belonging to the tile
        width: the width of the tile interior (excluding borders)
        left_border_width: the width of the left border
        right_border_width: the width of the right border
    """

    left_most_tile_pixel: int
    width: int
    left_border_width: int
    right_border_width: int


@dataclass
class TileBorder2D:
    """A 2D border.

    Attributes:
        left: the left border size
        top: the top border size
        right: the right border size
        bottom: the bottom border size
    """

    left: int
    top: int
    right: int
    bottom: int


class Tile2D:
    """A 2D tile.

    Attributes:
        roi: the tile roi
        border: the tile border
    """

    center: Rectangle
    border: TileBorder2D
    roi: Rectangle

    def __init__(self, center: Union[Tuple[int, int, int, int], Rectangle], border: TileBorder2D) -> None:
        """Initializes a 2D tile.

        Args:
            center: the tile center
            border: the tile border
        """
        self.center = Rectangle(*center)
        self.border = border
        self.roi = Rectangle(
            x=self.center.x - self.border.left,
            y=self.center.y - self.border.top,
            w=self.center.w + self.border.left + self.border.right,
            h=self.center.h + self.border.top + self.border.bottom,
        )

    def __eq__(self, other: object) -> bool:
        """Implementation of the equal operator for a 2D tile."""
        # Taken from https://stackoverflow.com/questions/54801832/mypy-eq-incompatible-with-supertype-object
        # Also refer to https://mypy.readthedocs.io/en/stable/common_issues.html#incompatible-overrides
        if not isinstance(other, Tile2D):
            # If we return NotImplemented, Python will automatically try
            # running other.__eq__(self), in case 'other' knows what to do with Tile2D objects.
            return NotImplemented
        return self.center == other.center and self.border == other.border and self.roi == other.roi


class TilingStrategy2D(metaclass=ABCMeta):
    """Base module for creating a strategy to store the data samples on disk in a defined format"""

    @abstractmethod
    def tile_rectangle(self, rectangle: Union[Tuple[int, int, int, int], Rectangle]) -> List[Tile2D]:
        """Tiles the provided rectangle.

        Args:
            rectangle: The rectangle to tile.

        Returns:
            A list with the tiles covering the specified rectangle.
        """



# CZTile provides a set of tiling strategies
# Copyright 2022 Carl Zeiss Microscopy GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# To obtain a commercial version please contact Carl Zeiss Microscopy GmbH.
"""Provides tiling primitives for the DNN tiling strategy.

This is a python implementation of the algorithm described in README.
"""
import sys
from math import ceil, floor
from typing import List, Union, Tuple
from cztile.tiling_strategy import TilingStrategy2D, Rectangle, Tile2D, TileBorder2D, Tile1d


class AlmostEqualBorderFixedTotalAreaStrategy2D(TilingStrategy2D):
    """A 2D tiling strategy that covers a total area with a minimal number of tiles of constant
    total area such that:
        a) all interior tiles have at least a minimum border width/height on all sides
        b) the edge tiles have zero border at the edge and at least the minimum border width
           on their inner sides.
        c) The sizes of all non-zero borders differ at most by one pixel.

    Attributes:
        total_tile_width    The fixed total widths (inc. border) of the tiles to generate.
        total_tile_height   The fixed total heights (inc. border) of the tiles to generate.
        min_border_width    The minimum width of each of the interior borders.
    """

    def __init__(self, total_tile_width: int, total_tile_height: int, min_border_width: int) -> None:
        """Initializes an AlmostEqualBorderFixedTotalAreaStrategy2D.

        Args:
            total_tile_width: The fixed total widths (inc. border) of the tiles to generate.
            total_tile_height: The fixed total heights (inc. border) of the tiles to generate.
            min_border_width: The minimum width of each of the interior borders.
        """
        self.total_tile_width = total_tile_width
        self.total_tile_height = total_tile_height
        self.min_border_width = min_border_width

        self.horizontal_tiler = Tiler1d(total_tile_width=total_tile_width, min_border_width=min_border_width)
        self.vertical_tiler = Tiler1d(total_tile_width=total_tile_height, min_border_width=min_border_width)

    def tile_rectangle(self, rectangle: Union[Tuple[int, int, int, int], Rectangle]) -> List[Tile2D]:
        """Tiles the provided rectangle.

        Args:
            rectangle: The rectangle to tile.

        Returns:
            A list with the tiles covering the specified rectangle.
        """
        rectangle = Rectangle(*rectangle)
        if rectangle.w * rectangle.h == 0:
            return []

        horizontal_tiles = self.horizontal_tiler.calculate_tiles(rectangle.w)
        vertical_tiles = self.vertical_tiler.calculate_tiles(rectangle.h)

        result: List[Tile2D] = []
        for horizontal_tile in horizontal_tiles:
            for vertical_tile in vertical_tiles:
                tile_center = Rectangle(
                    x=rectangle.x + horizontal_tile.left_most_tile_pixel,
                    y=rectangle.y + vertical_tile.left_most_tile_pixel,
                    w=horizontal_tile.width,
                    h=vertical_tile.width,
                )
                tile_border = TileBorder2D(
                    left=horizontal_tile.left_border_width,
                    top=vertical_tile.left_border_width,
                    right=horizontal_tile.right_border_width,
                    bottom=vertical_tile.right_border_width,
                )
                result.append(Tile2D(center=tile_center, border=tile_border))

        return result


class Tiler1d:
    """Tiles an integer interval beginning at 0 with tiles having a fixed total width and with zero outer borders.

    Attributes:
        total_tile_width    The fixed total widths of the tiles to generate.
        min_border_width    The minimum width of each of the interior borders.
    """

    def __init__(self, total_tile_width: int, min_border_width: int) -> None:
        """Initializes a 1D tiler.

        Args:
            total_tile_width: The fixed total widths of the tiles to generate.
            min_border_width: The minimum width of each of the interior borders.

        Raises:
            ValueError: For invalid total tile or min border width
        """
        if total_tile_width <= 0:
            raise ValueError("Total tile width must be greater than zero.")
        if min_border_width < 0:
            raise ValueError("Minimum border width cannot be negative.")
        if min_border_width > sys.maxsize / 2:
            raise ValueError(f"Minimum border width is too large. It needs to be smaller than {sys.maxsize / 2}.")

        self.total_tile_width = total_tile_width
        self.min_border_width = min_border_width
        if 2 * min_border_width >= total_tile_width:
            raise ValueError("Minimum border width must be less than half the tile size.")

    def calculate_tiles(self, image_width: int) -> List[Tile1d]:
        """Calculates a tiling of the specified image width.

        Args:
            image_width: The width of the image.

        Returns:
            A list of tiles covering the specified image_width.

        Raises:
            AssertionError: Border misconfigurations
        """
        if image_width < self.total_tile_width:
            border_width = self.total_tile_width - image_width
            left_border = border_width // 2
            right_border = border_width - left_border
            return [Tile1d(0, image_width, left_border, right_border)]

        if image_width == self.total_tile_width:
            return [Tile1d(0, self.total_tile_width, 0, 0)]

        max_width_of_edge_tiles: int = self.total_tile_width - self.min_border_width
        max_width_of_interior_tiles: int = self.total_tile_width - 2 * self.min_border_width
        interior: int = max(0, image_width - 2 * max_width_of_edge_tiles)

        number_of_tiles: int = ceil(interior * 1.0 / max_width_of_interior_tiles) + 2
        number_of_non_zero_borders: int = 2 * number_of_tiles - 2

        excess_border: float = (
            2.0 * max_width_of_edge_tiles + (number_of_tiles - 2) * max_width_of_interior_tiles - image_width
        )

        fractional_excess_border: float = excess_border / number_of_non_zero_borders
        fractional_border_width: float = fractional_excess_border + self.min_border_width

        cumulative_border: List[int] = [0] * (number_of_non_zero_borders + 1)
        for j in range(0, number_of_non_zero_borders + 1):
            cbj: float = j * fractional_border_width
            cumulative_border[j] = floor(cbj)

        tile_boundaries: List[int] = [0] * (number_of_tiles + 1)
        tile_boundaries[0] = 0
        tile_boundaries[number_of_tiles] = image_width
        for i in range(1, number_of_tiles):
            tile_boundaries[i] = i * self.total_tile_width - cumulative_border[2 * i - 1]

        result: List[Tile1d] = []
        for k in range(0, number_of_tiles):
            left_most_tile_pixel = tile_boundaries[k]
            width = tile_boundaries[k + 1] - tile_boundaries[k]
            total_border = self.total_tile_width - width

            if k == 0:
                left_border = 0
            elif k == number_of_tiles - 1:
                left_border = total_border
            else:
                left_border_index = 2 * k
                left_border = cumulative_border[left_border_index] - cumulative_border[left_border_index - 1]

                # Some assertions. These can be removed without harm.
                right_border1 = total_border - left_border
                right_border2 = cumulative_border[left_border_index + 1] - cumulative_border[left_border_index]
                if right_border1 != right_border2:
                    raise AssertionError("right_border1 != right_border2")
                if right_border1 < self.min_border_width:
                    raise AssertionError("right border < self.min_border_width")
                if left_border < self.min_border_width:
                    raise AssertionError("left border < self.min_border_width")

            right_border = total_border - left_border

            if left_border + width + right_border != self.total_tile_width:
                raise AssertionError("LB + W + RB != Total Width")

            result.append(Tile1d(left_most_tile_pixel, width, left_border, right_border))

        return result



"""Holds all relevant information for packaging and publishing to PyPI."""
from typing import List

import setuptools

# no external requirements of now
requirements: List[str] = []

VERSION = "0.1.2"

# pylint: disable=line-too-long
with open("README.md", "r", encoding="utf-8") as fh_read:
    long_description = fh_read.read()
setuptools.setup(
    name="cztile",
    version=VERSION,
    author="Nuno Mesquita",
    author_email="nuno.mesquita@zeiss.com",
    description="A set of tiling utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # Note: Exclude test folder in MANIFEST.in to also remove from source dist
    # See https://stackoverflow.com/questions/8556996/setuptools-troubles-excluding-packages-including-data-files
    # See https://docs.python.org/3.6/distutils/sourcedist.html
    packages=setuptools.find_packages(exclude=["test", "test.*"]),
    license_files=["LICENSE.txt", "NOTICE.txt"],
    # Classifiers help users find your project by categorizing it.
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    # Make required Python version compliant with official TF docs (https://www.tensorflow.org/install)
    # It follows: We build a pure Python wheel
    # Note that Python used to build the sources specifies the Python version of the dist (relevant during build only)
    # See https://packaging.python.org/guides/distributing-packages-using-setuptools/#pure-python-wheels for more info
    # We also restrict the code to >3.6 to:
    # - fully benefit from type annotations (See https://realpython.com/python-type-checking/ for more info)
    # - having dataclasses natively supported (starting 3.7+)
    # - + see https://realpython.com/python-data-classes/
    # - + no need to include it as ext. pkg through backport in 3.6 (dev.to/hanpari/dataclasses-in-python-3-6-29id)
    python_requires=">3.6,<3.12",
    install_requires=requirements,
)
